#ifndef _FOLLY_MAP_TEST_H
#define _FOLLY_MAP_TEST_H

#include <folly/concurrency/ConcurrentHashMap.h>
#include <folly/AtomicHashMap.h>
#include <folly/AtomicUnorderedMap.h>

#include <cds_test/stress_test.h>
#include <cds_test/stress_test_util.h>

#include <algorithm>
#include <iostream>
#include <memory>
#include <random>
#include <thread>

namespace folly_test {

typedef folly::ConcurrentHashMap<size_t, size_t> ConcurrentHashMap;
typedef folly::AtomicHashMap<size_t, size_t> AtomicHashMap;
typedef folly::AtomicUnorderedInsertMap64<size_t, size_t>
    AtomicUnorderedInsertMap;

// AtomicUnorderedInsertMap
template <typename Key, typename Value>
bool map_insert(AtomicUnorderedInsertMap* map, Key key, Value value) {
  auto iter = map->find(key);
  if (iter == map->cend() || iter->second != value) {
    // Insert/update the <key,value> pair
    map->emplace(key, value);
    return true;
  } else {
    return false;
  }
}

// AtomicHashMap
template <typename Key, typename Value>
bool map_insert(AtomicHashMap* map, Key key, Value value) {
  auto iter = map->find(key);
  if (iter == map->end() || iter->second != value) {
    // Insert/update the <key,value> pair
    map->insert({key, value});
    return true;
  } else {
    return false;
  }
}

// ConcurrentHashMap
template <typename Key, typename Value>
bool map_insert(ConcurrentHashMap * map, Key key, Value value) {
  auto iter = map->find(key);
  if (iter == map->cend() || iter->second != value) {
    // Insert/update the <key,value> pair
    map->insert({key, value});
    return true;
  } else {
    return false;
  }
}

template <typename Map, typename Key>
bool map_find(const Map* map, Key key) {
  return map->find(key) != map->cend();
}

// Specialization for AtomicHashMap
template <typename Key>
bool map_find(const AtomicHashMap* map, Key key) {
  return map->find(key) != map->end();
}

// Doesn't erase for AtomicUnorderedInsertMap & AtomicHashMap, but returns
// whether the key exists in the map.
template <typename Map, typename Key>
bool map_delete(Map* map, Key key) {
  return map_find(map, key);
}

// Specialization for ConcurrentHashMap
template <typename Key>
bool map_delete(ConcurrentHashMap* map, Key key) {
  if (map_find(map, key)) {
    map->erase(key);
    return true;
  } else {
    return false;
  }
}

} // namespace folly_test

#endif
