#include "map_test.h"

namespace folly_test {

class FollyMapInsDelFindTest_Sequential : public cds_test::stress_fixture {
protected:
  static unsigned s_nInsertDeletePercentage;
  static size_t s_nMapSize;

  enum actions { do_find, do_insert, do_delete };
  static const unsigned int kShuffleSize = 100;
  static actions s_arrShuffle[kShuffleSize];

  static size_t s_nConcurrentHashMapPassCount;
  static size_t s_nAtomicHashMapPassCount;
  static size_t s_nAtomicUnorderedInsertMapPassCount;

  static void SetUpTestCase() {
    const cds_test::config& cfg = get_config("SequentialFollyMap");
    GetConfigNonZeroExpected(InsertDeletePercentage, 5);
    GetConfigNonZeroExpected(MapSize, 10000);
    GetConfigNonZeroExpected(ConcurrentHashMapPassCount, 1000);
    GetConfigNonZeroExpected(AtomicHashMapPassCount, 1000);
    GetConfigNonZeroExpected(AtomicUnorderedInsertMapPassCount, 1000);
  }

  template <typename Map> static void run_test(Map* map, size_t pass_count) {
    size_t nInsertedNum = 0;
    size_t nNotInsertedNum = 0;
    size_t nFindSuccess = 0;
    size_t nFindFailed = 0;
    size_t nDeletedNum = 0;
    size_t nOperations = 0;

    // The number to operate on the map.
    size_t n = s_nMapSize;
    for (size_t count = 0; count < pass_count; count++) {
      // Insert
      unsigned mod = count % kShuffleSize;
      if (mod < s_nInsertDeletePercentage) {
        if (map_insert(map, n, n)) {
          nInsertedNum++;
        } else if (map_insert(map, n, n + 1)) {
          nInsertedNum++;
        } else {
          nNotInsertedNum++;
        }
      } else {
        nNotInsertedNum++;
      }
      // Find
      if (map_find(map, n)) {
        ++nFindSuccess;
      } else {
        ++nFindFailed;
      }
      // Delete
      if (mod < s_nInsertDeletePercentage && map_delete(map, n)) {
        nDeletedNum++;
      }
      if (++n == 2 * s_nMapSize) {
        n = s_nMapSize;
      }
    }
    EXPECT_EQ(nInsertedNum, nDeletedNum);
    EXPECT_EQ(nInsertedNum, nFindSuccess);
    EXPECT_EQ(nFindFailed, nNotInsertedNum);
  }
};

size_t FollyMapInsDelFindTest_Sequential::s_nMapSize;
unsigned FollyMapInsDelFindTest_Sequential::s_nInsertDeletePercentage;
const unsigned int FollyMapInsDelFindTest_Sequential::kShuffleSize;
FollyMapInsDelFindTest_Sequential::actions
    FollyMapInsDelFindTest_Sequential::s_arrShuffle
        [FollyMapInsDelFindTest_Sequential::kShuffleSize];

size_t FollyMapInsDelFindTest_Sequential::s_nConcurrentHashMapPassCount;
size_t FollyMapInsDelFindTest_Sequential::s_nAtomicHashMapPassCount;
size_t FollyMapInsDelFindTest_Sequential::s_nAtomicUnorderedInsertMapPassCount;

TEST_F(FollyMapInsDelFindTest_Sequential, FollyConcurrentHashMap) {
  std::unique_ptr<ConcurrentHashMap> map(
      new ConcurrentHashMap(s_nMapSize));
  run_test(map.get(), s_nConcurrentHashMapPassCount);
}

TEST_F(FollyMapInsDelFindTest_Sequential, FollyAtomicHashMap) {
  std::unique_ptr<AtomicHashMap> map(
      new AtomicHashMap(s_nMapSize));
  run_test(map.get(), s_nAtomicHashMapPassCount);
}

TEST_F(FollyMapInsDelFindTest_Sequential, FollyAtomicUnorderedInsertMap) {
  std::unique_ptr<AtomicUnorderedInsertMap> map(
      new AtomicUnorderedInsertMap(s_nMapSize));
  run_test(map.get(), s_nAtomicUnorderedInsertMapPassCount);
}

} // namespace folly_test
